program dotNetApp;

//Compiler rejects attempts to unbox an object to a Boolean, which is acceptable in C#
{.$define USE_UNBOXING}

uses
  System.Windows.Forms,
  System.Reflection, //needed for late binding to the COM server
  Interop.ComServer; //needed for early binding to the COM server

type
  TfrmRCW = class(Form)
    btnLateBound: Button;
    btnEarlyBound: Button;
    procedure btnLateBoundClick(Sender: TObject; Args: EventArgs);
    procedure btnEarlyBoundClick(Sender: TObject; Args: EventArgs);
  public
    constructor Create;
  end;

{ TfrmPInvoke }

constructor TfrmRCW.Create;
begin
  inherited;
  Width := 327;
  Height := 90;
  StartPosition := FormStartPosition.CenterScreen;
  FormBorderStyle := FormBorderStyle.FixedSingle;
  MinimizeBox := False;
  MaximizeBox := False;
  Text := '.NET App Using Win32 COM Object';
  btnLateBound := Button.Create;
  with btnLateBound do
  begin
    Text := 'Late Bound Access';
    SetBounds(16, 16, 137, 33);
    Parent := Self;
    Add_Click(btnLateBoundClick);
  end;
  btnEarlyBound := Button.Create;
  with btnEarlyBound do
  begin
    Text := 'Early Bound Access';
    SetBounds(168, 16, 137, 33);
    Parent := Self;
    Add_Click(btnEarlyBoundClick);
  end;
end;

const
  Val = 45;

procedure TfrmRCW.btnLateBoundClick(Sender: TObject;
  Args: EventArgs);
var
  T: System.Type;
  ComObj, Result: TObject;
  OneArg: array[0..0] of TObject;
  TwoArgs: array[0..1] of TObject;
  ParamModifiers: array[0..0] of ParameterModifier;
  Answer: Int32;
begin
  T := System.Type.GetTypeFromProgID('ComServer.ComObject');
  ComObj := Activator.CreateInstance(T);
  //Set up a call to a method with a value parameter
  OneArg[0] := 'Hello world';
  T.Invokemember('One', BindingFlags.InvokeMethod, nil, ComObj, OneArg);
  //Set up a call to a method with a value parameter and a reference parameter
  TwoArgs[0] := Int32(Val);
  Answer := System.Int32(0);
  TwoArgs[1] := Answer; //pass by reference
  ParamModifiers[0] := ParameterModifier.Create(2); //the modifier has items for each parameter
  ParamModifiers[0].Item[0] := False; //1st arg is by value
  ParamModifiers[0].Item[1] := True;  //2nd arg is by reference
  Result := T.InvokeMember('Two', BindingFlags.InvokeMethod, nil, ComObj,
    TwoArgs, ParamModifiers, nil, nil);
  Answer := Int32(TwoArgs[1]);
{$ifdef USE_UNBOXING}
  if System.Boolean(Result) then
{$else}
  if Result.ToString = System.Boolean.TrueString then
{$endif}
    MessageBox.Show(Format('Twice %d is %s', [Val, Answer.ToString]));
end;

procedure TfrmRCW.btnEarlyBoundClick(Sender: TObject;
  Args: EventArgs);
var
  ComObj: IComInterface;
  Answer: Integer;
begin
  ComObj := ComObjectClass.Create as IComInterface;
  ComObj.One('Hello world');
  if ComObj.Two(Val, Answer) then
    MessageBox.Show(Format('Twice %d is %d', [Val, Answer]));
end;

begin
  Application.Run(TfrmRCW.Create);
end.
